# -*- coding: utf-8 -*-
"""
3D Object Property Manager
-------------------------------

Description
===========

This module contains methods and data-structures to manage all properties of objects (points, lines, sheeets and
solids) within the AEDT 3D Modeler

"""
from __future__ import absolute_import

import random
import string
from collections import defaultdict
from .. import generate_unique_name, retry_ntimes, aedt_exception_handler
from .GeometryOperators import GeometryOperators

clamp = lambda n, minn, maxn: max(min(maxn, n), minn)

rgb_color_codes = {
    "Black": (0, 0, 0),
    "Green": (0, 128, 0),
    "White": (255, 255, 255),
    "Red": (255, 0, 0),
    "Lime": (0, 255, 0),
    "Blue": (0, 0, 255),
    "Yellow": (255, 255, 0),
    "Cyan": (0, 255, 255),
    "Magenta": (255, 0, 255),
    "Silver": (192, 192, 192),
    "Gray": (128, 128, 128),
    "Maroon": (128, 0, 0),
    "Olive": (128, 128, 0),
    "Purple": (128, 0, 128),
    "Teal": (0, 128, 128),
    "Navy": (0, 0, 128),
    "copper": (184, 115, 51),
    "stainless steel": (224, 223, 219)
}
@aedt_exception_handler
def _uname(name=None):
    """Appends a 6-digit hash code to a specified name

    Parameters
    ----------
    name : str, default='NewObject_'
        Specified name

    Returns
    -------
    str
    """
    char_set = string.ascii_uppercase + string.digits
    unique_name = ''.join(random.sample(char_set, 6))
    if name:
        return name + unique_name
    else:
        return 'NewObject_' + unique_name


@aedt_exception_handler
def _to_boolean(val):
    """Get the boolean value of the provided input.

        If the value is a boolean return the value.
        Otherwise check to see if the value is in
        ["false", "f", "no", "n", "none", "0", "[]", "{}", "" ]
        and returns True if value is not in the list

    Parameters
    ----------
    val : bool or str
        Input value to be tested for True/False condition

    Returns
    -------
    bool
    """

    if val is True or val is False:
        return val

    false_items = ["false", "f", "no", "n", "none", "0", "[]", "{}", ""]

    return not str(val).strip().lower() in false_items


@aedt_exception_handler
def _dim_arg(value, units):
    """Concatenate a specified units string to a numerical input

    Parameters
    ----------
    value : str or number
        Valid expression string in the AEDT modeler, e.g. "5mm"
    units : str
        Valid units string in the AEDT modeler, e.g. "mm"

    Returns
    -------
    str
    """
    try:
        val = float(value)
        return str(val) + units
    except:
        return value



class VertexPrimitive(object):
    """Vertex object within the AEDT Desktop Modeler

    Parameters
    ----------
    parent : Object3d
        Pointer to the calling object which provides additional functionality

    id : int
        Object id as determined by the parent object

    """
    def __init__(self, parent, id):
        self.id = id
        self._parent = parent
        self._oeditor = parent.m_Editor

    @property
    def position(self):
        """Position of the vertex

        Returns
        -------
        list of float
            List of [x, y, z] coordinates of the vertex (in model units)
        """
        return [float(i) for i in list(self._oeditor.GetVertexPosition(self.id))]

    @aedt_exception_handler
    def chamfer(self, left_distance=1, right_distance=None, angle=45, chamfer_type=0):
        """Add chamfer to selected edge

        Parameters
        ----------
        left_distance : float, default=1.0
            left distance from the edge
        right_distance : float, default=None
            right distance from the edge
        angle : float, default=45.0
            angle value (for chamfer-type 2 and 3)
        chamfer_type : int, default=0
                * 0 - Symmetric
                * 1 - Left Distance-Right Distance
                * 2 - Left Distance-Angle
                * 3 - Right Distance-Angle

        Returns
        -------
        bool
            True if operation is successful
        """
        if self._parent.is3d:
            self._parent.messenger.add_error_message("chamfer is possible only on Vertex in 2D Objects ")
            return False
        vArg1 = ['NAME:Selections', 'Selections:=', self._parent.name, 'NewPartsModelFlag:=', 'Model']
        vArg2 = ["NAME:ChamferParameters"]
        vArg2.append('Edges:='), vArg2.append([])
        vArg2.append('Vertices:='), vArg2.append([self.id])
        vArg2.append('LeftDistance:='), vArg2.append(self._parent._parent.arg_with_dim(left_distance))
        if not right_distance:
            right_distance = left_distance
        if chamfer_type == 0:
            vArg2.append('RightDistance:='), vArg2.append(self._parent._parent.arg_with_dim(right_distance))
            vArg2.append('ChamferType:='), vArg2.append('Symmetric')
        elif chamfer_type == 1:
            vArg2.append('RightDistance:='), vArg2.append(self._parent._parent.arg_with_dim(right_distance))
            vArg2.append('ChamferType:='), vArg2.append('Left Distance-Right Distance')
        elif chamfer_type == 2:
            vArg2.append('Angle:='), vArg2.append(str(angle) + "deg")
            vArg2.append('ChamferType:='), vArg2.append('Left Distance-Right Distance')
        elif chamfer_type == 3:
            vArg2.append('Angle:='), vArg2.append(str(angle) + "deg")
            vArg2.append('ChamferType:='), vArg2.append('Right Distance-Angle')
        else:
            self._parent.messenger.add_error_message("Wrong Type Entered. Type must be integer from 0 to 3")
            return False
        self._parent.m_Editor.Chamfer(vArg1, ["NAME:Parameters", vArg2])
        if self._parent.name in list(self._parent.m_Editor.GetObjectsInGroup("UnClassified")):
            self._parent.odesign.Undo()
            self._parent.messenger.add_error_message("Operation Failed generating Unclassified object. Check and retry")
            return False
        return True

    @aedt_exception_handler
    def fillet(self, radius=0.1, setback=0.0):
        """Add Fillet to selected edge

        Parameters
        ----------
        radius : float, default=0.1
            Fillet Radius
        setback : float, default=0.0
            Fillet setback value

        Returns
        -------
        bool
            True if operation is successful
        """
        if self._parent.is3d:
            self._parent.messenger.add_error_message("chamfer is possible only on Vertex in 2D Objects ")
            return False
        vArg1 = ['NAME:Selections', 'Selections:=', self._parent.name, 'NewPartsModelFlag:=', 'Model']
        vArg2 = ["NAME:FilletParameters"]
        vArg2.append('Edges:='), vArg2.append([])
        vArg2.append('Vertices:='), vArg2.append([self.id])
        vArg2.append('Radius:='), vArg2.append(self._parent._parent.arg_with_dim(radius))
        vArg2.append('Setback:='), vArg2.append(self._parent._parent.arg_with_dim(setback))
        self._parent.m_Editor.Fillet(vArg1, ["NAME:Parameters", vArg2])
        if self._parent.name in list(self._parent.m_Editor.GetObjectsInGroup("UnClassified")):
            self._parent.odesign.Undo()
            self._parent.messenger.add_error_message("Operation Failed generating Unclassified object. Check and retry")
            return False
        return True

    def __repr__(self):
        return "Vertex "+str(self.id)

    def __str__(self):
        return "Vertex "+str(self.id)


class EdgePrimitive(object):
    """Edge object within the AEDT Desktop Modeler

    Parameters
    ----------
    parent : Object3d
        Pointer to the calling object which provides additional functionality

    id : int
        Object id as determined by the parent object

    """
    def __init__(self, parent, id):
        self.id = id
        self._parent = parent
        self._oeditor = parent.m_Editor
        self.vertices = []
        for vertex in self._oeditor.GetVertexIDsFromEdge(self.id):
            vertex = int(vertex)
            self.vertices.append(VertexPrimitive(parent, vertex))

    @property
    def midpoint(self):
        """Midpoint coordinates of the edge.
        If the edge is not a segment with two vertices return an empty list.

        Returns
        -------
        list of float
            Midpoint coordinates as [x, y, z] coordinates
        """

        if len(self.vertices) == 2:
            midpoint = GeometryOperators.get_mid_point(self.vertices[0].position, self.vertices[1].position)
            return list(midpoint)
        elif len(self.vertices) == 1:
            return self.vertices[0].position
        else:
            return []

    @property
    def length(self):
        """Length of the edge

        Returns
        -------
        float or False
            Edge length in model units, False if edge does not have 2 vertices
        """
        if len(self.vertices) == 2:
            length = GeometryOperators.points_distance(self.vertices[0].position, self.vertices[1].position)
            return float(length)
        else:
            return False

    @aedt_exception_handler
    def chamfer(self, left_distance=1, right_distance=None, angle=45, chamfer_type=0):
        """Add Chamfer to selected edge

        Parameters
        ----------
        left_distance : float, default=1.0
            left distance from the edge
        right_distance : float, default=None
            right distance from the edge
        angle : float, default=45.0
            angle value (for chamfer-type 2 and 3)
        chamfer_type : int, default=0
                * 0 - Symmetric
                * 1 - Left Distance-Right Distance
                * 2 - Left Distance-Angle
                * 3 - Right Distance-Angle

        Returns
        -------
        bool
            True if operation is successful
        """

        vArg1 = ['NAME:Selections', 'Selections:=', self._parent.name, 'NewPartsModelFlag:=', 'Model']
        vArg2 = ["NAME:ChamferParameters"]
        vArg2.append('Edges:='), vArg2.append([self.id])
        vArg2.append('Vertices:='), vArg2.append([])
        vArg2.append('LeftDistance:='), vArg2.append(self._parent._parent.arg_with_dim(left_distance))
        if not right_distance:
            right_distance=left_distance
        if chamfer_type == 0:
            vArg2.append('RightDistance:='), vArg2.append(self._parent._parent.arg_with_dim(right_distance))
            vArg2.append('ChamferType:='), vArg2.append('Symmetric')
        elif chamfer_type == 1:
            vArg2.append('RightDistance:='), vArg2.append(self._parent._parent.arg_with_dim(right_distance))
            vArg2.append('ChamferType:='), vArg2.append('Left Distance-Right Distance')
        elif chamfer_type == 2:
            vArg2.append('Angle:='), vArg2.append(str(angle)+"deg")
            vArg2.append('ChamferType:='), vArg2.append('Left Distance-Right Distance')
        elif chamfer_type == 3:
            vArg2.append('Angle:='), vArg2.append(str(angle)+"deg")
            vArg2.append('ChamferType:='), vArg2.append('Right Distance-Angle')
        else:
            self._parent.messenger.add_error_message("Wrong Type Entered. Type must be integer from 0 to 3")
            return False
        self._oeditor.Chamfer(vArg1, ["NAME:Parameters", vArg2])
        if self._parent.name in list(self._parent.m_Editor.GetObjectsInGroup("UnClassified")):
            self._parent.odesign.Undo()
            self._parent.messenger.add_error_message("Operation Failed generating Unclassified object. Check and retry")
            return False
        return True

    @aedt_exception_handler
    def fillet(self, radius=0.1, setback=0):
        """Add Fillet to selected edge

        Parameters
        ----------
        radius : float, default=0.1
            Fillet radius
        setback : float, default=0.0
            Fillet setback

        Returns
        -------
        bool
            True if operation is successful

        """
        vArg1 = ['NAME:Selections', 'Selections:=', self._parent.name, 'NewPartsModelFlag:=', 'Model']
        vArg2 = ["NAME:FilletParameters"]
        vArg2.append('Edges:='), vArg2.append([self.id])
        vArg2.append('Vertices:='), vArg2.append([])
        vArg2.append('Radius:='), vArg2.append(self._parent._parent.arg_with_dim(radius))
        vArg2.append('Setback:='), vArg2.append(self._parent._parent.arg_with_dim(setback))
        self._parent.m_Editor.Fillet(vArg1, ["NAME:Parameters", vArg2])
        if self._parent.name in list(self._parent.m_Editor.GetObjectsInGroup("UnClassified")):
            self._parent.odesign.Undo()
            self._parent.messenger.add_error_message("Operation Failed generating Unclassified object. Check and retry")
            return False
        return True

    def __repr__(self):
        return "EdgeId "+str(self.id)

    def __str__(self):
        return "EdgeId "+str(self.id)


class FacePrimitive(object):
    """ """
    def __repr__(self):
        return "FaceId "+str(self.id)

    def __str__(self):
        return "FaceId "+str(self.id)

    def __init__(self, parent, id):
        self._id = id
        self._parent = parent
        self._oeditor = self._parent.m_Editor
        self.edges = []
        self.vertices = []
        for edge in self._oeditor.GetEdgeIDsFromFace(self.id):
            edge = int(edge)
            self.edges.append(EdgePrimitive(parent, edge))
        for vertex in self._oeditor.GetVertexIDsFromFace(self.id):
            vertex = int(vertex)
            self.vertices.append(VertexPrimitive(parent, vertex))

    @property
    def id(self):
        return self._id

    @property
    def center(self):
        """Get the face center (in model units)

        Returns
        -------
        list of float or False
            Center position in [x, y, z] coordinates - only works for planar faces. otherwise return False
        """
        try:
            c = self._parent.m_Editor.GetFaceCenter(self.id)
        except:
            self._parent.messenger.add_warning_message("Non Planar Faces doesn't provide any Face Center")
            return False
        center = [float(i) for i in c]
        return center

    @property
    def centroid(self):
        """Get the face center (in model units)

        Returns
        -------
        list of float
            Centroid of all vertices of the face.
        """
        return GeometryOperators.get_polygon_centroid([pos.position for pos in self.vertices])


    @property
    def area(self):
        """Get the face area

        Returns
        -------
        float
            Face area (in model units)
        """
        area = self._parent.m_Editor.GetFaceArea(self.id)
        return area

    @aedt_exception_handler
    def move_with_offset(self, offset=1.0):
        """Move face along normal

        Parameters
        ----------
        offset : float, default=1.0
            Offset to apply (in model units)

        Returns
        -------
        bool
            True if operation is successful
        """
        self._parent.m_Editor.MoveFaces(["NAME:Selections", "Selections:=", self._parent.name, "NewPartsModelFlag:=", "Model"],
                                    ["NAME:Parameters",
                                     ["NAME:MoveFacesParameters", "MoveAlongNormalFlag:=", True, "OffsetDistance:=", _dim_arg(offset, self._parent.object_units),
                                      "MoveVectorX:=", "0mm", "MoveVectorY:=", "0mm", "MoveVectorZ:=", "0mm",
                                      "FacesToMove:=", [self.id]]])
        return True

    @aedt_exception_handler
    def move_with_vector(self, vector):
        """Move face Along a specified vector

        Parameters
        ----------
        vector : list of float [x, y, z]
            Move vector to apply to the face

        Returns
        -------
        bool
            True if operation is successful

        """
        self._parent.m_Editor.MoveFaces(
                ["NAME:Selections", "Selections:=",  self._parent.name, "NewPartsModelFlag:=", "Model"],
                ["NAME:Parameters",
                 ["NAME:MoveFacesParameters", "MoveAlongNormalFlag:=", False, "OffsetDistance:=", "0mm",
                  "MoveVectorX:=", _dim_arg(vector[0], self._parent.object_units), "MoveVectorY:=",
                  _dim_arg(vector[1], self._parent.object_units), "MoveVectorZ:=",
                  _dim_arg(vector[2], self._parent.object_units),
                  "FacesToMove:=", [self.id]]])
        return True

    @property
    def normal(self):
        """Get the face normal.

        Limitations:
        #. the face must be planar.
        #. Currently it works only if the face has at least two vertices. Notable excluded items are circles and
        ellipses that have only one vertex.
        #. If a bounding box is specified, the normal is orientated outwards with respect to the bounding box.
        Usually the bounding box refers to a volume where the face lies.
        If no bounding box is specified, the normal can be inward or outward the volume.

        Returns
        -------
        list of float
            normal vector (normalized [x, y, z]) or None.

        """
        vertices_ids = self.vertices
        if len(vertices_ids) < 2 or not self.center:
            self._parent.messenger.add_warning_message("Not enough vertices or non-planar face")
            return None
        # elif len(vertices_ids)<2:
        #     v1 = vertices_ids[0].position
        #     offset = [(a-b)/2 for a,b in zip(v1,self.center)]
        #     v2 = [a+b for a,b in zip(self.center, offset)]
        else:
            v1 = vertices_ids[0].position
            v2 = vertices_ids[1].position
        fc = self.center
        cv1 = GeometryOperators.v_points(fc, v1)
        cv2 = GeometryOperators.v_points(fc, v2)
        if  cv2[0]==cv1[0]==0.0 and  cv2[1]==cv1[1]==0.0:
            n = [0,0,1]
        elif cv2[0]==cv1[0]==0.0 and  cv2[2]==cv1[2]==0.0:
            n = [0,1,0]
        elif cv2[1]==cv1[1]==0.0 and  cv2[2]==cv1[2]==0.0:
            n = [1,0,0]
        else:
            n = GeometryOperators.v_cross(cv1, cv2)
        normal = GeometryOperators.normalize_vector(n)

        """
        try to move the face center twice, the first with the normal vector, and the second with its inverse.
        Measures which is closer to the center point of the bounding box.
        """
        inv_norm = [-i for i in normal]
        mv1 = GeometryOperators.v_sum(fc, normal)
        mv2 = GeometryOperators.v_sum(fc, inv_norm)
        bb_center = GeometryOperators.get_mid_point(self._parent.bounding_box[0:3], self._parent.bounding_box[3:6])
        d1 = GeometryOperators.points_distance(mv1, bb_center)
        d2 = GeometryOperators.points_distance(mv2, bb_center)
        if d1 > d2:
            return normal
        else:
            return inv_norm


class Object3d(object):
    """Object Attributes Manager for the AEDT 3D Modeler

    Basic usage demonstrated with an HFSS design:

    ##################################################
    >>> from pyaedt import Hfss
    >>> aedtapp = Hfss()
    >>> prim = aedtapp.modeler.primitives


    Then we can create a part. Create_box returns an Object3d object

    ##################################################
    >>> id = prim.create_box([0, 0, 0], [10, 10, 5], "Mybox", "Copper")
    >>> part = prim[id]

    Properties
    * bounding_box
    * faces
    * edges
    * vertices
    * material_name
    * object_units

    """
    def __init__(self, parent, name=None):
        if name:
            self._m_name = name
        else:
            self._m_name = _uname()
        self._parent = parent
        self.flags = ""
        self._transparency = 0.3
        self._part_coordinate_system = "Global"
        self._solve_inside = True
        self._wireframe = False
        self._model = True
        self._color = (132, 132, 193)
        self._bounding_box = None
        self._object_type = None
        self._material_name = self._parent.defaultmaterial
        self._solve_inside = None

    @property
    def analysis_type(self):
        """Returns True for a 'Sheet' object in a 2D analysis or a 'Solid' object in a 3D analysis"""
        solve_3d = self._parent.is3d and self.object_type == "Solid"
        solve_2d = not self._parent.is3d and self.object_type == "Sheet"
        return solve_3d or solve_2d

    @property
    def bounding_box(self):
        """Return the bounding box of the individual part

        List of six [x, y, z] positions of the bounding box containing
        Xmin, Ymin, Zmin, Xmax, Ymax, and Zmax values

        This is done by creating a new empty design, copying the object there and getting the model bounding box. A
        list of six 3-D position vectors is returned

        Returns
        -------
        list of [list of float]
        """
        if self.id >0:
            origin_design_name = self._parent._parent.design_name
            unique_design_name = generate_unique_name("bounding")
            new_design = self._parent.oproject.InsertDesign("HFSS", unique_design_name, "", "")
            self.m_Editor.Copy(["NAME:Selections", "Selections:=", self.name])
            oeditor = new_design.SetActiveEditor("3D Modeler")
            oeditor.Paste()
            bb = list(oeditor.GetModelBoundingBox())
            self._bounding_box = [float(b) for b in bb]
            self._parent._parent.delete_design(name=unique_design_name, fallback_design=origin_design_name)
        return self._bounding_box

    @property
    def odesign(self):
        return self._parent.odesign

    @property
    def faces(self):
        """ Returns the a the information for each face in the given part

        Returns
        -------
        list of FacePrimitive
        """
        faces = []
        for face in self.m_Editor.GetFaceIDs(self.name):
                face = int(face)
                faces.append(FacePrimitive(self, face))
        return faces

    @property
    def top_face(self):
        """The top face in the Z direction of the object.

        Returns
        -------
        FacePrimitive
        """
        result = [(float(face.center[2]), face) for face in self.faces]
        result = sorted(result, key=lambda tup: tup[0])
        return result[-1][1]

    @property
    def bottom_face(self):
        """The bottom face in the Z direction of the object.

        Returns
        -------
        FacePrimitive
        """
        result = [(float(face.center[2]), face) for face in self.faces]
        result = sorted(result, key=lambda tup: tup[0])
        return result[0][1]

    @property
    def edges(self):
        """Returns the information for each edge in the given part

        Returns
        -------
        list of EdgePrimitive
        """
        edges = []
        for edge in self._parent.get_object_edges(self.name):
                edge = int(edge)
                edges.append(EdgePrimitive(self, edge))
        return edges

    @property
    def vertices(self):
        """Returns the information for each vertex in the given part

        Returns
        -------
        list of VertexPrimitive
        """
        vertices = []
        for vertex in self._parent.get_object_vertices(self.name):
                vertex = int(vertex)
                vertices.append(VertexPrimitive(self, vertex))
        return vertices

    @property
    def m_Editor(self):
        """Provides a pointer to the oEditor object in the AEDT API. Intended primarily for use by
           FacePrimitive, EdgePrimitive, VertexPrimitive child objects

        Returns
        -------
        oEditor COM Object
        """
        return self._parent.oeditor

    @property
    def messenger(self):
        """Provides a pointer to the oEditor object in the AEDT API. Intended primarily for use by
           FacePrimitive, EdgePrimitive, VertexPrimitive child objects

        Returns
        -------
        AEDTMessageManager
        """
        return self._parent.messenger

    @property
    def material_name(self):
        """Get or set the material name of the object
        If the material does not exist, then send a warning and do nothing
        """
        return self._material_name

    @material_name.setter
    def material_name(self, mat):
        if self._parent.materials.checkifmaterialexists(mat):
            self._material_name = mat
            vMaterial = ["NAME:Material", "Material:=", mat]
            self._change_property(vMaterial)
            self._material_name = mat
        else:
            self.messenger.add_warning_message("Material {} does not exist".format(mat))

    @property
    def id(self):
        """Object id
        """
        return self._parent.oeditor.GetObjectIDByName(self._m_name)

    @property
    def object_type(self):
        """Get the object type
            Returns str:
             * Solid
             * Sheet
             * Line
         """
        return self._object_type

    @property
    def is3d(self):
        if self.object_type == "Solid":
            return True
        else:
            return False

    @property
    def name(self):
        """Get or set the object name as a string value
        """
        return self._m_name

    @name.setter
    def name(self, obj_name):
        if obj_name != self._m_name:
            vName = []
            vName.append("NAME:Name")
            vName.append("Value:=")
            vName.append(obj_name)
            self._change_property(vName)
            self._m_name = obj_name


    @property
    def color(self):
        """Get or set part color as a tuple of integer values for (Red, Green, Blue)
        If the integer values are outside the range 0 - 255, then limit the values. Invalid inputs will be ignored
         >>> part.color = (255,255,0)
        """
        return self._color

    @property
    def color_string(self):
        """Return the color tuple as a string in the format '(Red, Green, Blue)'"""
        return "({} {} {})".format(self.color[0], self.color[1], self.color[2])

    @color.setter
    def color(self, color_value):
        color_tuple = None
        if isinstance(color_value, str):
            try:
                color_tuple = rgb_color_codes[color_value]
            except KeyError:
                parse_string = color_value.replace(')', '').replace('(', '').split()
                if len(parse_string) == 3:
                    color_tuple = tuple([int(x) for x in parse_string])
        else:
            color_tuple = color_value

        if color_tuple:

            try:
                R = clamp(color_tuple[0], 0, 255)
                G = clamp(color_tuple[1], 0, 255)
                B = clamp(color_tuple[2], 0, 255)
                vColor = ["NAME:Color", "R:=", str(R), "G:=", str(G), "B:=", str(B)]
                self._change_property(vColor)
                self._color = (R, G, B)
            except TypeError:
                color_tuple = None

        if not color_tuple:
            msg_text = "Invalid color input {} for object {}".format(color_value, self._m_name)
            self._parent.messenger.add_warning_message(msg_text)


    @property
    def transparency(self):
        """Get or set part transparency as a value between 0.0 and 1.0
        If the value is outside the range, then apply a limit. If the value is not a valid number, set to 0.0
        """
        return self._transparency

    @transparency.setter
    def transparency(self, T):
        try:
            trans_float = float(T)
            if trans_float < 0.0:
                trans_float = 0.0
            elif trans_float > 1.0:
                trans_float = 1.0
        except ValueError:
            trans_float = 0.0
        vTrans = ["NAME:Transparent", "Value:=", str(trans_float)]

        self._change_property(vTrans)

        self._transparency = trans_float
        return True

    @property
    def object_units(self):
        return self._parent.model_units

    @property
    def part_coordinate_system(self):
        """Get or set the part coordinate systems

        Returns
        -------
        str
            Name of the part coordinate system
        """
        return self._part_coordinate_system

    @part_coordinate_system.setter
    def part_coordinate_system(self, sCS):
        self._part_coordinate_system = sCS
        return True

    @property
    def solve_inside(self):
        """Get or Set part solve inside flag as a boolean value
        True if "solve-inside" is activated for the part
        """
        return self._solve_inside

    @solve_inside.setter
    def solve_inside(self, S):
        vSolveInside = []
        # fS = self._to_boolean(S)
        fs = S
        vSolveInside.append("NAME:Solve Inside")
        vSolveInside.append("Value:=")
        vSolveInside.append(fs)

        self._change_property(vSolveInside)

        self._solve_inside = fs

    @aedt_exception_handler
    def export_attributes_legacy(self, name=None):
        if name:
            obj_name = name
        else:
            obj_name = self.name

        args = ["NAME:Attributes", "Name:=", obj_name, "Flags:=", self.flags, "Color:=", self.color_string,
                "Transparency:=", self.transparency, "PartCoordinateSystem:=", self.part_coordinate_system,
                "MaterialName:=", self.material_name, "SolveInside:=", self.solve_inside]

        return args

    @aedt_exception_handler
    def export_attributes(self, name=None):
        if name:
            obj_name = name
        else:
            obj_name = self.name

        args = ["NAME:Attributes",
                "Name:=", obj_name,
                "Flags:=", self.flags,
                "Color:=", self.color_string,
                "Transparency:=", self.transparency,
                "PartCoordinateSystem:=", self.part_coordinate_system,
                "UDMId:=", "",
                "MaterialValue:=", chr(34) + self._material_name + chr(34),
                "SurfaceMaterialValue:=", chr(34) +"Steel-oxidised-surface"+ chr(34),
                "SolveInside:=", self.solve_inside]

        if self._parent.version >= "2021.1":
            args += ["ShellElement:="	, False,
                     "ShellElementThickness:=", "0mm",
                     "IsMaterialEditable:=", True,
                     "UseMaterialAppearance:=", False,
                     "IsLightweight:=", False]

        return args

    @property
    def display_wireframe(self):
        """Get or set the display_wireframe property of the part as a boolean value
        """
        return self._wireframe

    @display_wireframe.setter
    def display_wireframe(self, fWireframe):
        vWireframe = ["NAME:Display Wireframe", "Value:=", fWireframe]
        # fwf = self._to_boolean(wf)

        self._change_property(vWireframe)
        self._wireframe = fWireframe

    @property
    def model(self):
        """Set part Model/Non-model as a boolean value
        """
        return self._model

    @model.setter
    def model(self, fModel):
        vArg1 = ["NAME:Model", "Value:=", fModel]
        fModel = _to_boolean(fModel)
        self._change_property(vArg1)
        self._model = fModel

    @aedt_exception_handler
    def clone(self):
        """Clones the object and returns the Object3d object

        Returns
        -------
        Object3d
        """
        new_obj_tuple = self._parent.modeler.clone(self.id)
        success = new_obj_tuple[0]
        assert success, "Could not clone the object {}".format(self.name)
        new_name = new_obj_tuple[1]
        new_object = self._parent[new_name]
        return new_object

    @aedt_exception_handler
    def _change_property(self, vPropChange):
        """

        Parameters
        ----------
        vPropChange :


        Returns
        -------

        """
        vChangedProps = ["NAME:ChangedProps", vPropChange]

        #TODO potential for name conflict
        vPropServers = ["NAME:PropServers", self._m_name]
        # ScriptArgInf saPropServers(args_propservers)
        # saPropServers.ExportVariant(vPropServers)

        vGeo3d = ["NAME:Geometry3DAttributeTab", vPropServers, vChangedProps]

        vOut = ["NAME:AllTabs", vGeo3d]
        if vPropChange == ['NAME:Solve Inside', 'Value:=', True]:
            print(1)
        self.m_Editor.ChangeProperty(vOut)
        return True

    @aedt_exception_handler
    def update_object_type(self):
        if self._m_name in self._parent.solids:
            self._object_type = "Solid"
        else:
            if self._m_name in self._parent.sheets:
                self._object_type = "Sheet"
            elif self._m_name in self._parent.lines:
                self._object_type = "Line"

    @aedt_exception_handler
    def update_properties(self):

        n = 10
        all_prop = retry_ntimes(n, self.m_Editor.GetProperties, "Geometry3DAttributeTab", self._m_name)

        if 'Solve Inside' in all_prop:
            solveinside = retry_ntimes(n, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, 'Solve Inside')
            if solveinside == 'false' or solveinside == 'False':
                self._solve_inside = False
        if 'Material' in all_prop:
            mat = retry_ntimes(n, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, 'Material')
            if mat:
                self._material_name = mat[1:-1].lower()
            else:
                self._material_name = ''
        if 'Orientation' in all_prop:
            self.part_coordinate_system = retry_ntimes(n, self.m_Editor.GetPropertyValue,
                                                    "Geometry3DAttributeTab", self._m_name, 'Orientation')
        if 'Model' in all_prop:
            mod = retry_ntimes(n, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, 'Model')
            if mod == 'false' or mod == 'False':
                self._model = False
            else:
                self._model = True
        if 'Group' in all_prop:
            self.m_groupName = retry_ntimes(n, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, 'Group')
        if 'Display Wireframe' in all_prop:
            wireframe = retry_ntimes(n, self.m_Editor.GetPropertyValue,
                                     "Geometry3DAttributeTab", self._m_name, 'Display Wireframe')
            if wireframe == 'true' or wireframe == 'True':
                self._wireframe = True
        if 'Transparent' in all_prop:
            transp = retry_ntimes(n, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, 'Transparent')
            try:
                self._transparency = float(transp)
            except:
                self._transparency = 0.3
        if 'Color' in all_prop:
            color = int(retry_ntimes(n, self.m_Editor.GetPropertyValue, "Geometry3DAttributeTab", self._m_name, 'Color'))
            if color:
                b = (color >> 16) & 255
                g = (color >> 8) & 255
                r = color & 255
                self._color = (r, g, b)
            else:
                self._color = (0, 195, 255)
        if 'Surface Material' in all_prop:
            self.m_surfacematerial = retry_ntimes(n, self.m_Editor.GetPropertyValue,
                                               "Geometry3DAttributeTab", self._m_name, 'Surface Material')

class Padstack(object):
    """ """

    def __init__(self, name="Padstack", padstackmanager=None, units="mm"):
        self.name = name
        self.padstackmgr = padstackmanager
        self.units = units
        self.lib = ""
        self.mat = "copper"
        self.plating = 100
        self.layers = defaultdict(self.PDSLayer)
        self.hole = self.PDSHole()
        self.holerange = "UTL"
        self.solder_shape = "None"
        self.solder_placement = "abv"
        self.solder_rad = "0mm"
        self.sb2 = "0mm"
        self.solder_mat = "solder"
        self.layerid = 1

    class PDSHole(object):
        """ """
        def __init__(self, holetype="Cir", sizes=["1mm"], xpos="0mm", ypos="0mm", rot="0deg"):
            self.shape = holetype
            self.sizes = sizes
            self.x = xpos
            self.y = ypos
            self.rot = rot

    class PDSLayer(object):
        """ """

        def __init__(self, layername="Default", id=1):
            self.layername = layername
            self.id = id
            self._pad = None
            self._antipad = None
            self._thermal = None
            self.connectionx = 0
            self.connectiony = 0
            self.connectiondir = 0

        @property
        def pad(self):
            """ """
            return self._pad

        @property
        def antipad(self):
            """ """
            return self._antipad

        @property
        def thermal(self):
            """ """
            return self._thermal

        @pad.setter
        def pad(self, value=None):
            """

            Parameters
            ----------
            value :
                 (Default value = None)

            Returns
            -------

            """
            if value:
                self._pad = value
            else:
                self._pad = self.PDSHole(holetype="None", sizes=[])

        @property
        def antipad(self):
            """ """
            return self._antipad

        @antipad.setter
        def antipad(self, value=None):
            """

            Parameters
            ----------
            value :
                 (Default value = None)

            Returns
            -------

            """
            if value:
                self._antipad = value
            else:
                self._antipad =  self.PDSHole(holetype="None", sizes=[])

        @property
        def thermal(self):
            """ """
            return self._thermal

        @thermal.setter
        def thermal(self, value=None):
            """

            Parameters
            ----------
            value :
                 (Default value = None)

            Returns
            -------

            """
            if value:
                self._thermal = value
            else:
                self._thermal = self.PDSHole(holetype="None", sizes=[])

        class PDSHole:
            """ """
            def __init__(self, holetype="Cir", sizes=["1mm"], xpos="0mm", ypos="0mm", rot="0deg"):
                self.shape = holetype
                self.sizes = sizes
                self.x = xpos
                self.y = ypos
                self.rot = rot
    @property
    def pads_args(self):
        """ """
        arg = ["NAME:" + self.name, "ModTime:=", 1594101963, "Library:=", "", "ModSinceLib:=", False,
               "LibLocation:=", "Project"]
        arg2 = ["NAME:psd", "nam:=", self.name, "lib:=", "", "mat:=", self.mat,
                "plt:=", self.plating]
        arg3 = ["NAME:pds"]
        for el in self.layers:
            arg4 = []
            arg4.append("NAME:lgm")
            arg4.append("lay:=")
            arg4.append(self.layers[el].layername)
            arg4.append("id:=")
            arg4.append(el)
            arg4.append("pad:=")
            arg4.append(["shp:=", self.layers[el].pad.shape, "Szs:=", self.layers[el].pad.sizes, "X:=",
                         self.layers[el].pad.x, "Y:=", self.layers[el].pad.y, "R:=", self.layers[el].pad.rot])
            arg4.append("ant:=")
            arg4.append(["shp:=", self.layers[el].antipad.shape, "Szs:=", self.layers[el].antipad.sizes, "X:=",
                         self.layers[el].antipad.x, "Y:=", self.layers[el].antipad.y, "R:=",
                         self.layers[el].antipad.rot])
            arg4.append("thm:=")
            arg4.append(["shp:=", self.layers[el].thermal.shape, "Szs:=", self.layers[el].thermal.sizes, "X:=",
                         self.layers[el].thermal.x, "Y:=", self.layers[el].thermal.y, "R:=",
                         self.layers[el].thermal.rot])
            arg4.append("X:=")
            arg4.append(self.layers[el].connectionx)
            arg4.append("Y:=")
            arg4.append(self.layers[el].connectiony)
            arg4.append("dir:=")
            arg4.append(self.layers[el].connectiondir)
            arg3.append(arg4)
        arg2.append(arg3)
        arg2.append("hle:=")
        arg2.append(
            ["shp:=", self.hole.shape, "Szs:=", self.hole.sizes, "X:=", self.hole.x, "Y:=", self.hole.y, "R:=",
             self.hole.rot])
        arg2.append("hRg:=")
        arg2.append(self.holerange)
        arg2.append("sbsh:=")
        arg2.append(self.solder_shape)
        arg2.append("sbpl:=")
        arg2.append(self.solder_placement)
        arg2.append("sbr:=")
        arg2.append(self.solder_rad)
        arg2.append("sb2:=")
        arg2.append(self.sb2)
        arg2.append("sbn:=")
        arg2.append(self.solder_mat)
        arg.append(arg2)
        arg.append("ppl:=")
        arg.append([])
        return arg

    @aedt_exception_handler
    def add_layer(self, layername="Start", pad_hole=None, antipad_hole=None, thermal_hole=None, connx=0, conny=0,
                  conndir=0):
        """create a new layer in padstack

        Parameters
        ----------
        layername :
            Name of layer (Default value = "Start")
        pad_hole :
            pad hole object. Create it with add_hole command (Default value = None)
        antipad_hole :
            antipad hole object. Create it with add_hole command (Default value = None)
        thermal_hole :
            thermal hole object. Create it with add_hole command (Default value = None)
        connx :
            connection x direction (Default value = 0)
        conny :
            connection y direction (Default value = 0)
        conndir :
            connection attach angle (Default value = 0)

        Returns
        -------
        type
            True if succeeded

        """
        if layername in self.layers:
            return False
        else:
            new_layer = self.PDSLayer(layername, self.layerid)
            self.layerid += 1
            new_layer.pad = pad_hole
            new_layer.antipad = antipad_hole
            new_layer.thermal = thermal_hole
            new_layer.connectionx = connx
            new_layer.connectiony = conny
            new_layer.connectiondir = conndir
            self.layers[layername] = new_layer
            return True

    @aedt_exception_handler
    def add_hole(self, holetype="Cir", sizes=[1], xpos=0, ypos=0, rot=0):
        """

        Parameters
        ----------
        holetype :
            No" // no pad
            "Cir" // Circle
            "Sq" // Square
            "Rct" // Rectangle
            "Ov" // Oval
            "Blt" // Bullet
            "Ply" // Polygons
            "R45" // Round 45 thermal
            "R90" // Round 90 thermal
            "S45" // Square 45 thermal
            "S90" // Square 90 thermal (Default value = "Cir")
        sizes :
            array of sizes. Depends on the object. eg. Circle is an array of 1 element (Default value = [1])
        xpos :
            xposition (Default value = 0)
        ypos :
            yposition of hole (Default value = 0)
        rot :
            rotation angle (Default value = 0)

        Returns
        -------
        type
            hole object to be passed to padstack or layer

        """
        hole = self.PDSHole()
        hole.shape = holetype
        sizes =[_dim_arg(i, self.units) for i in sizes if type(i) is int or float]
        hole.sizes = sizes
        hole.x = _dim_arg(xpos, self.units)
        hole.y = _dim_arg(ypos, self.units)
        hole.rot = _dim_arg(rot, "deg")
        return hole

    @aedt_exception_handler
    def create(self):
        """Create Padstack in AEDT

        :return:

        Parameters
        ----------

        Returns
        -------

        """
        self.padstackmgr.Add(self.pads_args)
        return True

    @aedt_exception_handler
    def update(self):
        """Update Padstack in AEDT

        :return:

        Parameters
        ----------

        Returns
        -------

        """
        self.padstackmgr.Edit(self.name, self.pads_args)

    @aedt_exception_handler
    def remove(self):
        """REmove Padstack in AEDT

        :return:

        Parameters
        ----------

        Returns
        -------

        """
        self.padstackmgr.Remove(self.name, True, "", "Project")


class CircuitComponent(object):
    """description of class"""

    @property
    def composed_name(self):
        """ """
        if self.id:
            return self.name + ";" +str(self.id) + ";" + str(self.schematic_id)
        else:
            return self.name + ";" + str(self.schematic_id)

    def __init__(self, editor=None, units="mm", tabname="PassedParameterTab"):
        self.name = None
        self.m_Editor = editor
        self.modelName = None
        self.status = "Active"
        self.component = None
        self.id = 0
        self.refdes = ""
        self.schematic_id = 0
        self.levels = 0.1
        self.angle = 0
        self.x_location = "0mil"
        self.y_location = "0mil"
        self.mirror = False
        self.usesymbolcolor = True
        self.units = "mm"
        self.tabname = tabname
        self.InstanceName = None

    @aedt_exception_handler
    def set_location(self, x_location=None, y_location=None):
        """Set part location

        Parameters
        ----------
        x_location :
            x location (Default value = None)
        y_location :
            y location (Default value = None)

        Returns
        -------

        """
        if x_location is None:
            x_location = self.x_location
        else:
            x_location = _dim_arg(x_location,"mil")
        if y_location is None:
            y_location = self.y_location
        else:
            y_location = _dim_arg(y_location,"mil")

        vMaterial = ["NAME:Component Location", "X:=", x_location,"Y:=", y_location]
        self.change_property(vMaterial)

    @aedt_exception_handler
    def set_angle(self, angle=None):
        """Set part angle

        Parameters
        ----------
        angle :
            angle (Default value = None)

        Returns
        -------

        """
        if not angle:
            angle = str(self.angle) +"°"
        else:
            angle = str(angle) +"°"
        vMaterial = ["NAME:Component Angle", "Value:=", angle]
        self.change_property(vMaterial)

    @aedt_exception_handler
    def set_mirror(self, mirror_value=None):
        """Mirror part

        Parameters
        ----------
        mirror_value :
            mirror angle (Default value = None)

        Returns
        -------

        """
        if not mirror_value:
            mirror_value = self.mirror
        vMaterial = ["NAME:Component Mirror", "Value:=", mirror_value]
        self.change_property(vMaterial)

    @aedt_exception_handler
    def set_use_symbol_color(self, symbol_color=None):
        """Set Symbol Color usage

        Parameters
        ----------
        symbol_color :
            Bool (Default value = None)

        Returns
        -------

        """
        if not symbol_color:
            symbol_color = self.usesymbolcolor
        vMaterial = ["NAME:Use Symbol Color", "Value:=", symbol_color]
        self.change_property(vMaterial)
        return True


    @aedt_exception_handler
    def set_color(self, R=255, G=128, B=0):
        """Set Symbol Color

        Parameters
        ----------
        R :
            red (Default value = 255)
        G :
            Green (Default value = 128)
        B :
            Blue (Default value = 0)

        Returns
        -------

        """
        self.set_mirror(True)
        vMaterial = ["NAME:Component Color", "R:=", R,"G:=", G, "B:=", B]
        self.change_property(vMaterial)
        return True

    @aedt_exception_handler
    def set_property(self, property_name, property_value):
        """Set part property

        Parameters
        ----------
        property_name :
            property name
        property_value :
            property value

        Returns
        -------

        """
        if type(property_name) is list:
            for p,v in zip(property_name, property_value):
                v_prop = ["NAME:"+p, "Value:=", v]
                self.change_property(v_prop)
                self.__dict__[p] = v
        else:
            v_prop = ["NAME:" + property_name, "Value:=", property_value]
            self.change_property(v_prop)
            self.__dict__[property_name] = property_value
        return True

    @aedt_exception_handler
    def _add_property(self, property_name, property_value):
        """

        Parameters
        ----------
        property_name :

        property_value :


        Returns
        -------

        """
        self.__dict__[property_name] = property_value
        return True

    def change_property(self, vPropChange, names_list=None):
        """

        Parameters
        ----------
        vPropChange :

        names_list :
             (Default value = None)

        Returns
        -------

        """
        vChangedProps = ["NAME:ChangedProps", vPropChange]
        if names_list:
            vPropServers = ["NAME:PropServers"]
            for el in names_list:
                vPropServers.append(el)
        else:
            vPropServers = ["NAME:PropServers", self.composed_name]
        tabname = None
        if vPropChange[0][5:] in list(self.m_Editor.GetProperties(self.tabname, self.composed_name)):
            tabname = self.tabname
        elif vPropChange[0][5:] in list(self.m_Editor.GetProperties("PassedParameterTab", self.composed_name)):
            tabname = "PassedParameterTab"
        elif vPropChange[0][5:] in list(self.m_Editor.GetProperties("BaseElementTab", self.composed_name)):
            tabname = "BaseElementTab"
        if tabname:
            vGeo3dlayout = ["NAME:"+tabname, vPropServers, vChangedProps]
            vOut = ["NAME:AllTabs", vGeo3dlayout]
            return retry_ntimes(10, self.m_Editor.ChangeProperty, vOut)
        return False


class Objec3DLayout(object):
    """Class managing Objects Properties in HFSS 3DLauoit"""
    def __init__(self, parent):
        self._parent = parent
        self._n = 10

    @property
    def m_Editor(self):
        """ """
        return self._parent.oeditor

    @property
    def object_units(self):
        """ """
        return self._parent.model_units

    @aedt_exception_handler
    def change_property(self, vPropChange, names_list=None):
        """

        Parameters
        ----------
        vPropChange :

        names_list :
             (Default value = None)

        Returns
        -------

        """
        vChangedProps = ["NAME:ChangedProps", vPropChange]
        if names_list:
            vPropServers = ["NAME:PropServers"]
            for el in names_list:
                vPropServers.append(el)
        else:
            vPropServers = ["NAME:PropServers", self.name]
        vGeo3dlayout = ["NAME:BaseElementTab", vPropServers, vChangedProps]
        vOut = ["NAME:AllTabs", vGeo3dlayout]

        self.m_Editor.ChangeProperty(vOut)
        return True

    @aedt_exception_handler
    def set_property_value(self, property_name, property_value):
        """Set Property Value

        Parameters
        ----------
        property_name :
            property name
        property_value :
            property value

        Returns
        -------
        type
            Bool

        """
        vProp = ["NAME:"+property_name, "Value:=",property_value]
        return self.change_property(vProp)


class Components3DLayout(Objec3DLayout, object):
    """Class Containing Components in HFSS 3D Layout"""
    def __init__(self, parent, name=""):
        Objec3DLayout.__init__(self, parent)
        self.name = name

    @aedt_exception_handler
    def get_location(self):
        """:return:  Component location"""
        location = retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Location')
        return list(location)

    @aedt_exception_handler
    def get_placement_layer(self):
        """:return:  Component Placement Layer"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'PlacementLayer')

    @aedt_exception_handler
    def get_part(self):
        """:return:  Component Part"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Part')

    @aedt_exception_handler
    def get_part_type(self):
        """:return:  Component Part Type"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Part Type')

    @aedt_exception_handler
    def get_angle(self):
        """:return:  Component Angle"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Angle')


class Nets3DLayout(Objec3DLayout, object):
    """Class containing Netw in HFSS3D Layout"""
    def __init__(self, parent, name=""):
        Objec3DLayout.__init__(self, parent)
        self.name = name


class Pins3DLayout(Objec3DLayout, object):
    """Class containing Pins in HFSS3D Layout"""
    def __init__(self, parent, componentname="", pinname="", name=""):
        Objec3DLayout.__init__(self, parent)
        self.componentname = componentname
        self.pinname = pinname
        self.name = name

    @aedt_exception_handler
    def get_location(self):
        """:return:  pin location"""
        location = retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Location')
        return location

    @aedt_exception_handler
    def get_start_layer(self):
        """:return:  pin start layer"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Start Layer')

    @aedt_exception_handler
    def get_stop_layer(self):
        """:return:  pin stop layer"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Stop Layer')

    @aedt_exception_handler
    def get_holediam(self):
        """:return:  pin hole diameter"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'HoleDiameter')

    @aedt_exception_handler
    def get_angle(self):
        """:return:  pin angle"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Angle')


class Geometries3DLayout(Objec3DLayout, object):
    """Class containing Geometries in HFSS3D Layout"""
    def __init__(self, parent, name, id=0):
        Objec3DLayout.__init__(self, parent)
        self.name = name
        self.id = id

    @aedt_exception_handler
    def get_placement_layer(self):
        """:return:  object placement layer"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'PlacementLayer')

    @aedt_exception_handler
    def get_net_name(self):
        """:return:  object net name"""
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, 'Net')

    @aedt_exception_handler
    def get_property_value(self, propertyname):
        """

        Parameters
        ----------
        propertyname :
            property name

        Returns
        -------
        type
            object property value

        """
        return retry_ntimes(self._n, self.m_Editor.GetPropertyValue, "BaseElementTab", self.name, propertyname)

    @aedt_exception_handler
    def set_layer(self, layer_name):
        """Set object layer

        Parameters
        ----------
        layer_name :
            layer name

        Returns
        -------

        """
        vMaterial = ["NAME:PlacementLayer", "Value:=", layer_name]
        return self.change_property(vMaterial)

    @aedt_exception_handler
    def set_lock_position(self, lock_position=True):
        """Set lock position

        Parameters
        ----------
        lock_position :
            bool (Default value = True)

        Returns
        -------

        """
        vMaterial = ["NAME:LockPosition", "Value:=", lock_position]
        return self.change_property(vMaterial)

    @aedt_exception_handler
    def set_negative(self, negative=False):
        """Set negative

        Parameters
        ----------
        negative :
            bool (Default value = False)

        Returns
        -------

        """
        vMaterial = ["NAME:Negative", "Value:=", negative]
        return self.change_property(vMaterial)

    @aedt_exception_handler
    def set_net_name(self, netname=""):
        """Set net name

        Parameters
        ----------
        netname :
            net name (Default value = "")

        Returns
        -------

        """
        vMaterial = ["NAME:Net", "Value:=", netname]
        return self.change_property(vMaterial)




from __future__ import absolute_import
from ..generic.general_methods import aedt_exception_handler
from .Modeler import GeometryModeler
from  .Primitives3D import Primitives3D

class Modeler3D(GeometryModeler):
    """Modeler 3D class."""
    def __init__(self, application):
        GeometryModeler.__init__(self, application, is3d=True)
        self._primitives = Primitives3D(self._parent, self)
        self._primitivesDes = self._parent.project_name + self._parent.design_name

    def __get__(self, instance, owner):
        self._parent = instance
        return self

    @property
    def primitives(self):
        """ """
        if self._primitivesDes != self._parent.project_name + self._parent.design_name:
            self._primitives = Primitives3D(self._parent, self)
            self._primitivesDes = self._parent.project_name + self._parent.design_name
        return self._primitives

    @aedt_exception_handler
    def create_3dcomponent(self, component_file, component_name=None, variables_to_include=[], exclude_region=False):
        """Create 3D component file.

        Parameters
        ----------
        component_file : str
            Full path to the A3DCOMP file.
        component_name : str, optional
            Name of the component. The default is ``None``.
        variables_to_include : list
             List of variables to include. The default is ``[]``.
        exclude_region :
             Whether to exclude the region. The default is ``False``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if self._parent.design_type == "Icepak":
            exclude_region =True
        if not component_name:
            component_name = self._parent.design_name
        arg = ["NAME:CreateData", "ComponentName:=", component_name, "Company:=", "",
               "Company URL:=", "", "Model Number:=", "", "Help URL:=", "",
               "Version:=", "1.0", "Notes:=", "", "IconType:=", "",
               "Owner:=", "pyaedt", "Email:=", "", "Date:=", "9:44:15 AM  Mar 03, 2021",
               "HasLabel:=", False, "IsEncrypted:=", False, "AllowEdit:=", False,
               "SecurityMessage:=", "", "Password:=", "", "EditPassword:=", "",
               "PasswordType:=", "UnknownPassword", "HideContents:=", True,
               "ReplaceNames:=", True, "ComponentOutline:=", "None"]
        objs = self.primitives.get_all_objects_names()
        for el in objs:
            if "Region" in el and exclude_region:
                objs.remove(el)
        arg.append("IncludedParts:="), arg.append(objs)
        arg.append("HiddenParts:="), arg.append([])
        activecs = self.oeditor.GetActiveCoordinateSystem()
        allcs = self.oeditor.GetCoordinateSystems()
        arg.append("IncludedCS:="), arg.append(allcs)
        arg.append("ReferenceCS:="), arg.append(activecs)
        variables =variables_to_include
        arg.append("IncludedParameters:="), arg.append(variables)
        variables = self._parent._variable_manager.dependent_variable_names
        par_description = []
        for el in variables:
            par_description.append(el)
            par_description.append("")
        arg.append("IncludedDependentParameters:="), arg.append(variables)
        for el in variables:
            par_description.append(el)
            par_description.append("")
        arg.append("ParameterDescription:="), arg.append(par_description)
        arg.append("IsLicensed:="), arg.append(False)
        arg.append("LicensingDllName:="), arg.append("")
        arg.append("VendorComponentIdentifier:="), arg.append("")
        arg.append("PublicKeyFile:="), arg.append("")
        arg2 = ["NAME:DesignData"]
        boundaries = self.get_boundaries_name()
        arg2.append("Boundaries:="),arg2.append(boundaries)
        if self._parent.design_type == "Icepak":
            meshregions = [name for name in self._parent.mesh.meshregions.name]
            try:
                meshregions.remove("Global")
            except:
                pass
            arg2.append("MeshRegions:="), arg2.append(meshregions)
        else:
            excitations = self.get_excitations_name()
            arg2.append("Excitations:="), arg2.append(excitations)
        meshops = [el.name for el in self._parent.mesh.meshoperations]
        arg2.append("MeshOperations:="), arg2.append(meshops)
        arg3 = ["NAME:ImageFile", "ImageFile:=", ""]
        self.oeditor.Create3DComponent(arg, arg2, component_file, arg3)
        return True



"""
Material Library Class
----------------------------------------------------------------


Description
===========

This class contains all the functionalities to create and edit materials



"""
from __future__ import absolute_import
import os
import xml.etree.ElementTree as ET
from .Material import *
from ..generic.general_methods import aedt_exception_handler
import json

class Materials(object):
    """This class contains AEDT Materials database and methods for creating and manipulating materials
    """
    @property
    def odefinition_manager(self):
        """AEDT Definition Manager """
        return self._parent._oproject.GetDefinitionManager()

    @property
    def omaterial_manager(self):
        """AEDT Material Manager
        """
        return self.odefinition_manager.GetManager("Material")

    @property
    def messenger(self):
        """
        
        Returns
        -------
        MessageManager
            Message manager object
        """
        return self._parent._messenger

    @property
    def oproject(self):
        """AEDT Project object"""
        return self._parent.oproject

    def __init__(self, parent):
        self._parent = parent
        self.messenger.logger.info('Successfully loaded project materials !')
        self.material_keys = self._get_materials()
        self.surface_material_keys = self._get_surface_materials()
        self._load_from_project()
        pass

    def __len__(self):
        return len(self.material_keys)

    def __iter__(self):
        return self.material_keys.itervalues()

    def __getitem__(self, item):
        if item in list(self.material_keys.keys()):
            return self.material_keys[item]
        elif item in list(self.surface_material_keys.keys()):
            return self.surface_material_keys[item]


    @aedt_exception_handler
    def _get_materials(self):
        mats = {}
        try:
            for ds in self._parent.project_properies['AnsoftProject']['Definitions']['Materials']:
                mats[ds.lower()] = Material(self, ds.lower(), self._parent.project_properies['AnsoftProject']['Definitions']['Materials'][ds])
        except:
            pass
        return mats

    @aedt_exception_handler
    def _get_surface_materials(self):
        mats = {}
        try:
            for ds in self._parent.project_properies['AnsoftProject']['Definitions']['SurfaceMaterials']:
                mats[ds.lower()] = SurfaceMaterial(self, ds.lower(), self._parent.project_properies['AnsoftProject']['Definitions']['SurfaceMaterials'][ds])
        except:
            pass
        return mats


    @aedt_exception_handler
    def checkifmaterialexists(self, mat):
        """Check if a material Exists in AEDT. If it exists and it's not in material database it will populate it

        Parameters
        ----------
        mat : str
            material name

        Returns
        -------
        bool
            True/False

        """
        mat = mat.lower()
        lista = [i.lower() for i in list(self.odefinition_manager.GetProjectMaterialNames())]
        if mat in lista:
            return True
        else:
            mattry = self.omaterial_manager.GetData(mat)
            if mattry:
                self._aedmattolibrary(mat)
                return True
            else:
                return False

    @aedt_exception_handler
    def check_thermal_modifier(self, mat):
        """Check the termal modifier of predefined material. It checks in all the material properties to find any thermal modifiers

        Parameters
        ----------
        mat : str
            material name

        Returns
        -------
        bool

        """
        mat = mat.lower()
        exists = False
        if mat in self.material_keys:
            exists = True
        elif self.checkifmaterialexists(mat):
            self._load_from_project()
            exists = True
        if exists:
            omat = self.material_keys[mat]
            for el in MatProperties.aedtname:
                if omat.__dict__["_"+el].thermalmodifier:
                    return True
        return False

    @aedt_exception_handler
    def add_material(self, materialname):
        """The function create a new material named materialname and apply defaults value and return the material object
        that allows user to customize the material
        The function will not update the material in

        Parameters
        ----------
        materialname : str
            material name

        Returns
        -------
        Material
            Material Object

        Examples
        --------
        
        >>> from pyaedt import Hfss
        >>> hfss = Hfss()
        >>> mat = hfss.materials.add_material("MyMaterial")
        >>> print(mat.conductivity.value)
        
        """
        materialname = materialname.lower()
        self.messenger.add_info_message('Adding New Material material to Project Library: ' + materialname)
        if materialname in self.material_keys:
            self.messenger.add_warning_message(
                "Warning. The material is already in database. Please change name or edit it")
            return self.material_keys[materialname]
        else:
            material = Material(self._parent, materialname)
            material.update()
            self.messenger.add_info_message("Material added. Please edit it to update in Desktop")
            self.material_keys[materialname] = material
            return self.material_keys[materialname]

    @aedt_exception_handler
    def add_surface_material(self, material_name, emissivity=None):
        """The function create a new material surface in AEDT based properties are loaded from the XML file database amat.xml or from emissivity

        Parameters
        ----------
        material_name : str
            material name
        emissivity : float, optional
            Emissivity
        Returns
        -------
        SurfaceMaterial
            Material emissivity
            
        Examples
        --------
        
        >>> from pyaedt import Hfss
        >>> hfss = Hfss()
        >>> mat = hfss.materials.add_surface_material("Steel", 0.85)
        >>> print(mat.emissivity.value)
        """
        
        materialname = material_name.lower()
        self.messenger.add_info_message('Adding New Surface Material material to Project Library: ' + materialname)
        if materialname in self.surface_material_keys:
            self.messenger.add_warning_message(
                "Warning. The material is already in database. Please change name or edit it")
            return self.surface_material_keys[materialname]
        else:
            material = SurfaceMaterial(self._parent, materialname)
            if emissivity:
                material.emissivity = emissivity
                material.update()
            self.messenger.add_info_message("Material added. Please edit it to update in Desktop")
            self.surface_material_keys[materialname] = material
            return self.surface_material_keys[materialname]


    @aedt_exception_handler
    def _create_mat_project_vars(self, matlist):
        matprop={}
        tol = 1e-12
        for prop in MatProperties.aedtname:
            matprop[prop] = []
            for mat in matlist:
                try:
                    matprop[prop].append(float(mat.__dict__["_"+prop].value))
                except:
                    self.messenger.add_warning_message("Warning. Wrong parsed property. Reset to 0")
                    matprop[prop].append(0)
            try:
                a = sum(matprop[prop])
                if a <tol:
                    del matprop[prop]
            except:
                pass
        return matprop

    @aedt_exception_handler
    def add_material_sweep(self, swargs, matname):
        """The function create a new material named  made of an array of materials args
        If material needs a dataset (thermal modifier) than a dataset is created
        material args properties are loaded from the XML file database amat.xml
        enableTM: Boolean enable Thermal modifier in material description. At moment unused

        Parameters
        ----------
        swargs :
            list of materials to be merged into single sweep material
        matname :
            name of sweep material
        enableTM :
            bool (Default value = True)

        Returns
        -------
        int
            name of project variable index

        Examples
        --------
        
        >>> from pyaedt import Hfss
        >>> hfss = Hfss()
        >>> hfss.materials.add_material("MyMaterial")
        >>> hfss.materials.add_material("MyMaterial2")
        >>> hfss.materials.add_material_sweep(["MyMaterial", "MyMaterial2"], "Sweep_copper")
        """
        matsweep = []
        matname = matname.lower()
        for args in swargs:
            if args.lower() in [i.lower() for i in self.material_keys.keys()]:
                matsweep.append(self.material_keys[args.lower()])
            elif self.checkifmaterialexists(args):
                self._aedmattolibrary(args)
                matsweep.append(self.material_keys[args.lower()])

        mat_dict = self._create_mat_project_vars(matsweep)

        newmat = Material(self._parent, matname)
        index = "$ID"+matname
        self._parent[index] = 0
        for el in mat_dict:
            if el in list(mat_dict.keys()):
                self._parent["$"+matname+el] = mat_dict[el]
                newmat.__dict__["_"+el].value = "$"+matname+el+"["+ index + "]"
                newmat._update_props(el, "$"+matname+el+"["+ index + "]", False)

        newmat.update()
        self.material_keys[matname] = newmat
        return index

    @aedt_exception_handler
    def duplicate_material(self, material, new_name):
        """Duplicate Material

        Parameters
        ----------
        material : str
            material original name
        new_name : str
            target name

        Returns
        -------
        Material
            new material object

        Examples
        --------

        >>> from pyaedt import Hfss
        >>> hfss = Hfss()
        >>> hfss.materials.add_material("MyMaterial")
        >>> hfss.materials.duplicate_material("MyMaterial", "MyMaterial2")
        """
        if material.lower() not in list(self.material_keys.keys()):
            self.messenger.add_error_message("Material {} is not present".format(material))
            return False
        newmat = Material(self, new_name.lower(), self.material_keys[material.lower()]._props)
        newmat.update()
        self.material_keys[new_name.lower()] = newmat
        return newmat

    @aedt_exception_handler
    def duplicate_surface_material(self, material, new_name):
        """Duplicate Surface Material

        Parameters
        ----------
        material : str
            material original name
        new_name : str
            target name

        Returns
        -------
        SurfaceMaterial
            new Surface Material object

        Examples
        --------

        >>> from pyaedt import Hfss
        >>> hfss = Hfss()
        >>> hfss.materials.add_surface_material("MyMaterial")
        >>> hfss.materials.duplicate_surface_material("MyMaterial", "MyMaterial2")
        """
        if not material.lower() in list(self.surface_material_keys.keys()):
            self.messenger.add_error_message("Material {} is not present".format(material))
            return False
        newmat = SurfaceMaterial(self, new_name.lower(), self.surface_material_keys[material.lower()]._props)
        newmat.update()
        self.surface_material_keys[new_name.lower()] = newmat
        return newmat

    @aedt_exception_handler
    def remove_material(self, material, library="Project"):
        """Remove Material

        Parameters
        ----------
        material : str
            material name


        Returns
        -------
        bool
            ``True`` if succeeded


        Examples
        --------

        >>> from pyaedt import Hfss
        >>> hfss = Hfss()
        >>> hfss.materials.add_material("MyMaterial")
        >>> hfss.materials.remove_material("MyMaterial")

        """
        if material not in list(self.material_keys.keys()):
            self.messenger.add_error_message("Material {} is not present".format(material))
            return False
        self.odefinition_manager.RemoveMaterial(material, True, "", library)
        del self.material_keys[material]
        return True

    @property
    def conductors(self):
        """Get List of conductors in material database


        Returns
        -------
        list
            list of conductor names
        """
        data = []
        for key, mat in self.material_keys.items():
            if mat.is_conductor():
                data.append(key)
        return data

    @property
    def dielectrics(self):
        """Get List of dielectrics in material database


        Parameters
        ----------

        Returns
        -------
        list
            list of dielctric names
        """
        data = []
        for key, mat in self.material_keys.items():
            if mat.is_dielectric():
                data.append(key)
        return data

    def _load_from_project(self):
        mats = self.odefinition_manager.GetProjectMaterialNames()
        for el in mats:
            if el not in list(self.material_keys.keys()):
                try:
                    self._aedmattolibrary(el)
                except Exception as e:
                    self.messenger.add_info_message('aedmattolibrary failed for material {}'.format(el))

    @aedt_exception_handler
    def _aedmattolibrary(self, matname):
        matname = matname.lower()
        props = {}
        arg2dict(list(self.omaterial_manager.GetData(matname)), props)
        values_view = props.values()
        value_iterator = iter(values_view)
        first_value = next(value_iterator)
        newmat = Material(self, matname, first_value)

        self.material_keys[matname] = newmat
        return True

    def export_materials_to_file(self, full_json_path):
        """
        Save all materials to Json File

        Parameters
        ----------
        full_json_path : str
            full path to json file

        Returns
        -------
        bool
            ``True`` if succeeded
        """
        def find_datasets(d, out_list):
            for k, v in d.items():
                if isinstance(v, dict):
                    find_datasets(v, out_list)
                else:
                    if "pwl(" in str(v):
                        out_list.append(v[v.find("$"):v.find(",")])
        # Data to be written
        output_dict = OrderedDict()
        for el, val in self.material_keys.items():
            output_dict[el] = val._props
        out_list = []
        find_datasets(output_dict, out_list)
        datasets = OrderedDict()
        for ds in out_list:
            if ds in list(self._parent.project_datasets.keys()):
                d = self._parent.project_datasets[ds]
                if d.z:
                    datasets[ds] = OrderedDict({"Coordinates": OrderedDict({"DimUnits": [d.xunit, d.yunit, d.zunit],
                                                                            "Points": [val for tup in zip(d.x, d.y, d.z)
                                                                                       for val in tup]})})
                else:
                    datasets[ds] = OrderedDict({"Coordinates": OrderedDict(
                        {"DimUnits": [d.xunit, d.yunit], "Points": [val for tup in zip(d.x, d.y) for val in tup]})})
        json_dict = {}
        json_dict["materials"] = output_dict
        if datasets:
            json_dict["datasets"] = datasets

        with open(full_json_path, 'w') as fp:
            json.dump(json_dict, fp, indent=4)
        return True

    def import_materials_to_file(self, full_json_path):
        """
        Import and create materilas from json file

        Parameters
        ----------
        full_json_path : str
            full path to json file

        Returns
        -------
        bool
            ``True`` if succeeded
        """
        with open(full_json_path) as json_file:
            data = json.load(json_file)

        if "datasets" in list(data.keys()):
            for el, val in data["datasets"].items():
                numcol = len(val["Coordinates"]["DimUnits"])
                xunit = val["Coordinates"]["DimUnits"][0]
                yunit = val["Coordinates"]["DimUnits"][0]
                zunit = ""


                new_list = [val["Coordinates"]['Points'][i:i + numcol] for i in range(0, len(val["Coordinates"]['Points']), numcol)]
                xval = new_list[0]
                yval = new_list[1]
                zval = None
                if numcol>2:
                    zunit = val["Coordinates"]["DimUnits"][0]
                    zval = new_list[2]
                self._parent.create_dataset(el[1:],xunit=xunit, yunit=yunit, zunit=zunit, xlist=xval, ylist=yval, zlist=zval)

        for el, val in data["materials"].items():
            if el.lower() in list(self.material_keys.keys()):
                newname =generate_unique_name(el)
                self.messenger.add_warning_message("Material {} already exists. Renaming to {}".format(el, newname))
            else:
                newname = el
            newmat = Material(self, newname, val)
            newmat.update()
            self.material_keys[newname] = newmat
        return True




from .Analysis import Analysis
from ..modeler.Model2D import Modeler2D
from ..modules.Mesh import Mesh
from ..generic.general_methods import aedt_exception_handler, generate_unique_name


class FieldAnalysis2D(Analysis):
    """**AEDT_2D_FieldAnalysis**
    
    Class for 2D Field Analysis Setup (Maxwell2D, Q2D)
    
    It is automatically initialized by Application call. Refer to
    Application function for inputs definition

    Parameters
    ----------

    Returns
    -------

    """
    def __init__(self, application, projectname, designname, solution_type, setup_name=None,
                 specified_version=None, NG=False, AlwaysNew=False, release_on_exit=False):


        Analysis.__init__(self, application, projectname, designname, solution_type, setup_name,
                          specified_version, NG, AlwaysNew, release_on_exit)
        self._modeler = Modeler2D(self)
        self._mesh = Mesh(self)
        # self._post = PostProcessor(self)

    @property
    def modeler(self):
        """ """
        return self._modeler

    @property
    def mesh(self):
        """ """
        return self._mesh

    # @property
    # def post(self):
    #     return self._post

    @aedt_exception_handler
    def assignmaterial(self, obj, mat):
        """The function assigns Material mat to object obj. If material mat is not present it will be created

        Parameters
        ----------
        obj : str, list
            list of objects to which assign materials
        mat : str
            material to assign

        Returns
        -------
        type
            True if succeeded | False if failed

        """
        mat = mat.lower()
        selections = self.modeler.convert_to_selections(obj)
        arg1 = ["NAME:Selections"]
        arg1.append("Selections:="), arg1.append(selections)
        arg2 = ["NAME:Attributes"]
        arg2.append("MaterialValue:="), arg2.append(chr(34) + mat + chr(34))
        if mat in self.materials.material_keys:
            Mat = self.materials.material_keys[mat]
            Mat.update()
            if Mat.is_dielectric():
                arg2.append("SolveInside:="), arg2.append(True)
            else:
                arg2.append("SolveInside:="), arg2.append(False)
            self.modeler.oeditor.AssignMaterial(arg1, arg2)
            self._messenger.add_info_message('Assign Material ' + mat + ' to object ' + selections)
            if type(obj) is list:
                for el in obj:
                    self.modeler.primitives[el].material_name = mat
            else:
                self.modeler.primitives[obj].material_name = mat
            return True
        elif self.materials.checkifmaterialexists(mat):
            self.materials._aedmattolibrary(mat)
            Mat = self.materials.material_keys[mat]
            if Mat.is_dielectric():
                arg2.append("SolveInside:="), arg2.append(True)
            else:
                arg2.append("SolveInside:="), arg2.append(False)
            self.modeler.oeditor.AssignMaterial(arg1, arg2)
            self._messenger.add_info_message('Assign Material ' + mat + ' to object ' + selections)
            if type(obj) is list:
                for el in obj:
                    self.modeler.primitives[el].material_name = mat
            else:
                self.modeler.primitives[obj].material_name = mat

            return True
        else:
            self._messenger.add_error_message("Material Does Not Exists")
            return False



from ..generic.general_methods import aedt_exception_handler
from .Primitives import Primitives


class Primitives2D(Primitives, object):
    """Primitives2D class."""

    @aedt_exception_handler
    def is3d(self):
        """Returns False always to indicate a 3D analysis type."""
        return False

    @property
    def plane2d(self):
        """Create a 2D plane."""
        plane = "Z"
        if self._parent.design_type == "Maxwell 2D":
            if self._parent.odesign.GetGeometryMode()=="about Z":
                plane = "Y"
        return plane

    def __init__(self, parent, modeler):
        Primitives.__init__(self, parent, modeler)

    @aedt_exception_handler
    def create_circle(self, position, radius, numSides=0, name=None, matname=None):
        """Create a circle.

        Parameters
        ----------
        position :
            ApplicationName.modeler.Position(x,y,z) object
        radius : float
            Radius of the object.
        numSides : int, optional
            Number of sides. The default is ``0``, which is correct for a circle.
        name : str, optional
            Name of the object. The default is ``None``.
        matname : str, optional
            Name of the material. The default is ``None``. If ``None``,
            the default material is assigned.

        Returns
        -------
        type
            ID of the created circle.

        """
        o = self._new_object(matname=matname)

        szAxis = self.plane2d
        XCenter, YCenter, ZCenter = self.pos_with_arg(position)
        Radius = self.arg_with_dim(radius)
        o.material_name, o.solve_inside = self._check_material(matname, self.defaultmaterial)

        vArg1 = ["NAME:CircleParameters"]
        vArg1.append("XCenter:="), vArg1.append(XCenter)
        vArg1.append("YCenter:="), vArg1.append(YCenter)
        vArg1.append("ZCenter:="), vArg1.append(ZCenter)
        vArg1.append("Radius:="), vArg1.append(Radius)
        vArg1.append("WhichAxis:="), vArg1.append(szAxis)
        vArg1.append("NumSegments:="), vArg1.append('{}'.format(numSides))

        vArg2 = o.export_attributes(name)

        o._m_name =self.oeditor.CreateCircle(vArg1, vArg2)
        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_ellipse(self,position, major_raidus, ratio, bIsCovered=True, name=None, matname=None):
        """Create an ellipse.

        Parameters
        ----------
        position :
            ApplicationName.modeler.Position(x,y,z) object
        major_raidus : float
            
        ratio : float
            Ratio between major and minor radius.
        bIsCovered : bool, optional
            The default is ``True``.
        name : str, optional
            Name of the object. The default is ``None``.
        matname : str, optional
            Name of the material. The default is ``None``. If ``None``, 
            the default material is assigned.

        Returns
        -------
        type
           ID of the created ellipse.

        """
        o = self._new_object(matname=matname)

        szAxis = self.plane2d
        XStart, YStart, ZStart = self.pos_with_arg(position)

        MajorRadius = self.arg_with_dim(major_raidus)
        # Ratio = self.arg_with_dim(ratio)
        Ratio = ratio

        vArg1 = ["NAME:EllipseParameters"]
        vArg1.append("IsCovered:="), vArg1.append(bIsCovered)
        vArg1.append("XCenter:="), vArg1.append(XStart)
        vArg1.append("YCenter:="), vArg1.append(YStart)
        vArg1.append("ZCenter:="), vArg1.append(ZStart)
        vArg1.append("MajRadius:="), vArg1.append(MajorRadius)
        vArg1.append("Ratio:="), vArg1.append(Ratio)
        vArg1.append("WhichAxis:="), vArg1.append(szAxis)

        vArg2 = o.export_attributes(name)
        o._m_name =self.oeditor.CreateEllipse(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_rectangle(self, position, dimension_list, name=None, matname=None):
        """Create a rectangle.

        Parameters
        ----------
        position :
            ApplicationName.modeler.Position(x,y,z) object
        dimension_list : list
            dimension list
        name :
            Name of the object. The default is ``None``.
        matname : str, optional
            Name of the material. The default is ``None``. If ``None``, 
            the default material is assigned.

        Returns
        -------
        type
            ID of the created rectangle.

        """
        o = self._new_object(matname=matname)

        szAxis = self.plane2d
        XStart, YStart, ZStart = self.pos_with_arg(position)
        if self.plane2d == "Z":
            Height  = self.arg_with_dim(dimension_list[0])
            Width = self.arg_with_dim(dimension_list[1])
        else:
            Width = self.arg_with_dim(dimension_list[0])
            Height = self.arg_with_dim(dimension_list[1])

        vArg1 = ["NAME:RectangleParameters"]
        vArg1.append("XStart:="), vArg1.append(XStart)
        vArg1.append("YStart:="), vArg1.append(YStart)
        vArg1.append("ZStart:="), vArg1.append(ZStart)
        vArg1.append("Width:="), vArg1.append(Width)
        vArg1.append("Height:="), vArg1.append(Height)
        vArg1.append("WhichAxis:="), vArg1.append(szAxis)

        vArg2 = o.export_attributes(name)

        o._m_name =self.oeditor.CreateRectangle(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id



import time
from threading import Thread

import queue
from queue import Queue
import os.path
import subprocess
import clr
import os
if os.name == 'posix':
	import subprocessdotnet as subprocess
else:
	import subprocess
from .general_methods import env_path, env_value
class AedtSolve(object):
    '''
    class dedicated for calling Aedt solvers. Only solving on local machines is supported for the moment.
    Examples
    --------
    >>> from pyaedt import AedtSolve
    >>> solver = process.AedtSolve()
    >>> solver.NonGraphical = True
    >>> solver.ProjectPath = edb_file
    >>> solver.solve()
    '''
    def __init__(self,aedt_version="2021.1", aedt_installer_path=None):
        self._project_path = ""
        self._command = []
        self._nbcores = 4
        self._ng = True
        self._local = True
        self._logfile = ""
        if aedt_installer_path:
            self.installer_path = aedt_installer_path
        else:
            try:
                self.installer_path = env_path(aedt_version)
            except:
                raise Exception("Either a valid aedt version or full path has to be provided")

    @property
    def projectpath(self):
        return self._project_path

    @projectpath.setter
    def projectpath(self, value):
        self._project_path = value
        self._logfile = os.path.splitext(self._project_path)[0] + ".log"

    @property
    def exepath(self):
        return self._exe

    @exepath.setter
    def exepath(self,value):
        self._exe = value

    @property
    def command(self):
        return  self._command

    @command.setter
    def command(self,value):
        self._command = value

    @property
    def nbcores(self):
        return self._nbcores

    @nbcores.setter
    def nbcores(self,value):
        self._nbcores = value

    @property
    def nongraphical(self):
        return self._ng

    @nongraphical.setter
    def nongraphical(self,value):
        self._ng = value

    @property
    def local(self):
        return self._local

    @local.setter
    def local(self,value):
        self._local = value

    @property
    def logfile(self):
        return self._logfile

    @logfile.setter
    def logfile(self,value):
        self._logfile = value

    def solve(self):
        if os.name == "posix":
                self.command.append(os.path.join(self.installer_path,'ansysedt'))
        else:
               self.command.append(os.path.join(self.installer_path,'ansysedt.exe'))
        if self.nongraphical:
            self.command.append('-Batchsolve')
            self.command.append('-ng')
            self.command.append('-Monitor')
            self.command.append('-MachineList')
            # TODO Add batch option support
        if self._local:
            pass

        #self.Command.append('-Auto')
        self.command.append('-machinelist numcores={}'.format(self.nbcores))
        self.command.append('-LogFile')
        self.command.append(self.logfile)
        self.command.append(self.projectpath)
        print(self._command)
        p = subprocess.Popen(self.command)
        p.wait()

class SiwaveSolve(object):
    def __init__(self, aedt_version="2021.1", aedt_installer_path=None):
        self._project_path = ""
        self._exec_path = ""
        self._nbcores = 4
        self._ng = True
        if aedt_installer_path:
            self.installer_path = aedt_installer_path
        else:
            try:
                self.installer_path = env_path(aedt_version)
            except:
                raise Exception("Either a valid aedt version or full path has to be provided")


    @property
    def siwave_exe(self):
        return self._exe

    @siwave_exe.setter
    def siwave_exe(self,value):
        self._exe = value

    @property
    def projectpath(self):
        return self._project_path
        self._exec_path = os.path.splitext(self._project_path[0]) + ".exec"

    @projectpath.setter
    def projectpath(self,value):
        self._project_path = value

    @property
    def execfile(self):
        return self._exec_path

    @execfile.setter
    def execfile(self,value):
        self._exec_path = value

    @property
    def nbcores(self):
        return self._nbcores

    @nbcores.setter
    def nbcores(self,value):
        self._nbcores = value

    @property
    def nongraphical(self):
        return self._ng

    @nongraphical.setter
    def nongraphical(self,value):
        self._ng = value

    def solve(self):
        # supporting non graphical solve only
        if self.nongraphical:
            if os.name == "posix":
                exe_path = os.path.join(self.installer_path, 'siwave_ng')
            else:
                 exe_path = os.path.join(self.installer_path, 'siwave_ng.exe')     
            exec_file = os.path.splitext(self._project_path)[0] + '.exec'
            if os.path.exists(exec_file):
                with open(exec_file,"r+") as f:
                    content = f.readlines()
                    if not 'SetNumCpus' in content:
                        f.writelines(str.format('SetNumCpus {}', str(self.nbcores)) + '\n')
                        f.writelines("SaveSiw")
                    else:
                        fstarts = [i for i in range(len(content)) if content[i].startswith('SetNumCpus')]
                        content[fstarts] = str.format('SetNumCpus {}', str(self.nbcores))
                        f.closed()
                        os.remove(exec_file)
                        f = open(exec_file,"w")
                        f.writelines(content)
            command = [exe_path]
            command.append(self._project_path)
            command.append(exec_file)
            command.append('-formatOutput -useSubdir')
            p = subprocess.Popen(command)
            p.wait()



import os
from .generic.general_methods import aedt_exception_handler, generate_unique_name, retry_ntimes
from .hfss3dlayout import Hfss3dLayout
from .hfss import Hfss
from .circuit import Circuit
from .q3d import Q2d, Q3d
from .siwave import Siwave
from .icepak import Icepak
from .edb import Edb
from .maxwell import Maxwell2d, Maxwell3d
from .mechanical import Mechanical
from .rmxprt import Rmxprt
from .simplorer import Simplorer
from .desktop import Desktop
from .emit import Emit




"""
Message Manager Library Module
------------------------------

Description
===========
This class contains all the functionalities to log error and messages
both in AEDT and in log file.

Examples
--------
Log the three types of messages

>>> from pyaedt.hfss import Hfss
>>> hfss = Hfss()
>>> hfss._messenger.add_info_message("This is an info Message on Design", "Design")
>>> hfss._messenger.add_warning_message("This is a global warning Message", "Global")
>>> hfss._messenger.add_error_message("This is a Project Error Message", "Project")


"""

import sys
import logging
import os
from ..generic.general_methods import aedt_exception_handler, generate_unique_name
message_levels = {'Global': 0,
                  'Project': 1,
                  'Design': 2}

class Msg():
    (INFO, WARNING, ERROR, FATAL) = range(4)


class MessageList:
    """Data structure to collect and return message information from the AEDT Message Manager for a specified
       project name and design name

    Parameters
    ---------
    msg_list : list of str
        list of messages extracted by the AEDTMessenger class from the Desktop

    project_name : str, default=active project name
        Specified project within the AEDT Desktop
    design_name : str, default=active design name
        Specified design within the specified project


    Attributes
    ----------
    global_level : list of str
        List of strings representing the message content at the global level of the message manager

    project_level : list of str
        List of strings representing the message content of the specifiec project

    design_level : list of str
        List of strings representing the message content for the specified design within the specified project

    """
    def __init__(self, msg_list, project_name, design_name):
        self.global_level = []
        self.project_level = []
        self.design_level = []
        global_label = "Project: *Global - Messages, "
        project_label = "Project: {}, ".format(project_name)
        design_label = "Project: {}, Design: {} (".format(project_name, design_name)
        for line in msg_list:
            # Find the first instance of '[' to get the message context
            loc = line.find('[')
            if loc < 0:
                # Format is not clear - append to global
                self.global_level.append(line)
            else:
                line_label = line[0:loc]
                msg_txt = line[loc:].strip()
                if design_label in line_label:
                    self.design_level.append(msg_txt)
                elif project_label in line_label:
                    self.project_level.append(msg_txt)
                elif global_label in line_label:
                    self.global_level.append(msg_txt)


class AEDTMessageManager(object):
    """Class that manages AEDT Messaging to the logger file and to the AEDT Message UI

    Attributes
    ----------
    messages : MessageList

    Methods
    -------
    add_error message
    add_warning_message
    add_info_message
    add debug_message (deprecated)
    add message
    get_design_messages
    clear_messages

    """
    def __init__(self, parent=None, loadondesktop=True, loadonlog=True):
        self._parent = parent
        if not parent:
            self.MainModule = sys.modules['__main__']
            self._desktop = self.MainModule.oDesktop
        else:
            self._desktop = self._parent._desktop

        self.loadondesktop = loadondesktop
        if loadonlog:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = None

    @property
    def messages(self):
        """Return the AEDT Message Manager Contents for the active project and design

        Returns
        -------
        MessageList
        """
        return self.get_messages(self._project_name, self._design_name)

    @aedt_exception_handler
    def get_messages(self, project_name, design_name):
        """Return the AEDT Message Manager Contents for a specified project and design

        If the design- or project names are invalid, they will be ignored

        Returns
        -------
        MessageList

        """
        global_message_data = self._desktop.GetMessages("", "", 0)
        message_data = MessageList(global_message_data, project_name, design_name)
        return message_data

    @aedt_exception_handler
    def add_error_message(self, message_text, level=None):
        """Add a type 2 "Error" message to the message manager tree.

        Add an error message to the logger if the handler is present.

        Parameters
        ----------
        message_text :
            message to show
        level : str, default="Design"
            message level. Must be "Global", "Project" or "Design"

        Examples
        --------
        Add an error message to the desktop message manager

        >>> hfss._messenger.add_error_message("Project Error Message", "Project")

        """
        self.add_message(2, message_text, level)

    @aedt_exception_handler
    def add_warning_message(self, message_text, level=None):
        """Add a type 1 "Warning" message to the message manager tree.
        Add a warning message to the logger if the handler is present

        Parameters
        ----------
        message_text :
            message to show
        level : str, default="Design"
            message level. Must be "Global", "Project" or "Design"

        Examples
        --------
        Add a warning message to the desktop message manager

        >>> hfss._messenger.add_warning_message("Project warning message")

        """
        self.add_message(1, message_text, level)

    @aedt_exception_handler
    def add_info_message(self, message_text, level=None):
        """Add a type 0 "Info" message to the active design level of the message manager tree

        Add an info message to the logger if the handler is present

        Parameters
        ----------
        message_text : str
            Message to show
        level : str, default="Design"
            message level. Must be "Global", "Project" or "Design"

        Examples
        --------
        Add a info message at the global level.

        >>> hfss._messenger.add_info_message("Global warning message", "Global")

        """
        self.add_message(0, message_text, level)

    @aedt_exception_handler
    def add_debug_message(self, message_text, level=None):
        """
        .. deprecated use add_info_message
        """
        self.add_info_message(message_text, level)

    @aedt_exception_handler
    def add_message(self, type, message_text, level=None, proj_name=None, des_name=None):
        """Pass a parameterized message to the AEDT message manager allowing specification of type and project/design level.

        Parameters
        ----------
        type : int
            Message types:

            * 0: info
            * 1: Warning
            * 2: Error

        message_text : str
            Message to show.
        level : str, default="Design"
            message level. Must be "Global", "Project" or "Design"
        proj_name : str, optional
            Project name.
        des_name : str, optional
            Design name.

        """
        if not proj_name:
            proj_name = ""

        if not des_name:
            des_name = ""

        if not level:
            level = 'Design'

        assert level in message_levels, "Message level must be 'Design', 'Project', or 'Global'"

        if self.loadondesktop:
            if not proj_name and message_levels[level] > 0:
                proj_name = self._project_name
            if not des_name and message_levels[level] > 1:
                des_name = self._design_name
            if des_name and ";" in des_name:
                des_name = des_name[des_name.find(";")+1:]
            self._desktop.AddMessage(proj_name, des_name, type, message_text)

        if len(message_text) > 250:
            message_text = message_text[:250] + "..."

        # Print to stdout and to logger
        if type == 0:
            print("pyaedt Info: {}".format(message_text))
            self.logger.info(message_text)
        elif type == 1:
            print("pyaedt Warning: {}".format(message_text))
            self.logger.warning(message_text)
        elif type == 2:
            print("pyaedt Error: {}".format(message_text))
            self.logger.error(message_text)

    @aedt_exception_handler
    def clear_messages(self, proj_name=None, des_name=None, level=2):
        """Clear messages.

        Parameters
        ----------
        proj_name : str
            Name of project.  If blank, then all projects.  Defaults
            to current project.
        des_name : str
            Name of design If blank, then all designs.  Defaults
            to current project.
        level : int, optional
            Clear message level.  Default 2.

            * ``0`` : Clear all info messages
            * ``1`` : Clear all info and warning messages
            * ``2`` : Clear all info, warning and error messages
            * ``3`` : Clear all messages included info, warning,
              error, and fatal-error.

        Examples
        --------
        Clear all messages in the current design and project

        >>> hfss.clear_messages(level=3)

        """
        proj_name = self._project_name
        des_name = self._design_name
        self._desktop.ClearMessages(proj_name, des_name, level)

    @property
    def _oproject(self):
        return self._desktop.GetActiveProject()

    @property
    def _odesign(self):
        return self._oproject.GetActiveDesign()

    @property
    def _design_name(self):
        try:
            return self._odesign.GetName()
        except AttributeError:
            return ""

    @property
    def _project_name(self):
        try:
            return self._oproject.GetName()
        except AttributeError:
            return ""


class EDBMessageManager(object):
    """Class that manage EDB Messaging to the logger file"""

    def __init__(self, project_dir="C:\\Temp"):
        self.logger = logging.getLogger(__name__)
        if not self.logger.handlers:
            logging.basicConfig(
                filename=os.path.join(project_dir, "EDBTLib.log"),
                level=logging.DEBUG,
                format='%(asctime)s:%(name)s:%(levelname)-8s:%(message)s',
                datefmt='%Y/%m/%d %H.%M.%S')
            self.logger = logging.getLogger(__name__)

    def add_error_message(self, message_text):
        """Add a type 2 "Error" message to the logger

        Parameters
        ----------
        message_text :
            message to show

        Returns
        -------

        """

        self.logger.error(message_text)

    def add_warning_message(self, message_text):
        """Add a "Warning" message to the logger

        Parameters
        ----------
        message_text :
            message to show

        Returns
        -------

        """
        self.logger.warning(message_text)

    def add_info_message(self, message_text):
        """Add an "Info" message to the logger

        Parameters
        ----------
        message_text :
            message to show

        Returns
        -------

        """
        self.logger.info(message_text)



class SolutionType(object):
    """Default Solution Type Name"""
    class Hfss(object):
        """ """
        (DrivenModal, DrivenTerminal, EigenMode, Transient, SBR) = (
            "DrivenModal", "DrivenTerminal", "EigenMode", "Transient Network", "SBR+")

    class Maxwell3d(object):
        """ """
        (Transient, Magnetostatic, EddyCurrent, ElectroStatic, ElectroDCConduction, ElectroDCTransient) = \
            ("Transient", "Magnetostatic", "EddyCurrent", "Electrostatic", "ElectroDCConduction",
             "ElectricTransient")

    class Maxwell2d(object):
        """ """
        (TransientXY, TransientZ, MagnetostaticXY, MagnetostaticZ, EddyCurrentXY, EddyCurrentZ, ElectroStaticXY,
         ElectroStaticZ, ElectroDCConductionX, ElectroDCConductionZ, ElectroDCTransientXY, ElectroDCTransientZ) = \
            ("TransientXY", "TransientZ", "MagnetostaticXY", "MagnetostaticZ", "EddyCurrentXY", "EddyCurrentZ",
             "ElectrostaticXY", "ElectrostaticZ", "ElectroDCConductionXY","ElectroDCConductionZ",
             "ElectricTransientXY", "ElectricTransientZ")

    class Icepak(object):
        """ """
        (SteadyTemperatureAndFlow, SteadyTemperatureOnly, SteadyFlowOnly, TransientTemperatureAndFlow,
         TransientTemperatureOnly, TransientFlowOnly,) = (
            "SteadyTemperatureAndFlow", "SteadyTemperatureOnly", "SteadyFlowOnly", "TransientTemperatureAndFlow",
            "TransientTemperatureOnly", "TransientFlowOnly")

    class Circuit(object):
        """ """
        (NexximLNA, NexximDC, NexximTransient, NexximQuickEye, NexximVerifEye, NexximAMI, NexximOscillatorRSF,
         NexximOscillator1T, NexximOscillatorNT, NexximHarmonicBalance1T, NexximHarmonicBalanceNT, NexximSystem,
         NexximTVNoise, HSPICE, TR) = ("NexximLNA", "NexximDC", "NexximTransient",
                                       "NexximQuickEye", "NexximVerifEye", "NexximAMI", "NexximOscillatorRSF",
                                       "NexximOscillator1T",
                                       "NexximOscillatorNT", "NexximHarmonicBalance1T", "NexximHarmonicBalanceNT",
                                       "NexximSystem", "NexximTVNoise", "HSPICE", "TR")

    class Mechanical(object):
        """ """
        (Thermal, Structural, Modal) = ("Thermal", "Structural", "Modal")
class SetupTypes(object):
    """Default Setup Constants"""
    (HFSSDrivenAuto, HFSSDrivenDefault, HFSSEigen, HFSSTransient, HFSSSBR, MaxwellTransient, Magnetostatic, EddyCurrent,
     Electrostatic, ElectrostaticDC, ElectricTransient, SteadyTemperatureAndFlow, SteadyTemperatureOnly, SteadyFlowOnly,
     Matrix, NexximLNA, NexximDC, NexximTransient, NexximQuickEye, NexximVerifEye, NexximAMI, NexximOscillatorRSF,
     NexximOscillator1T, NexximOscillatorNT, NexximHarmonicBalance1T, NexximHarmonicBalanceNT, NexximSystem,
     NexximTVNoise, HSPICE, HFSS3DLayout, Open, Close, MechTerm, MechModal, GRM, TR, TransientTemperatureAndFlow,
     TransientTemperatureOnly, TransientFlowOnly,) = range(0, 39)


